#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scazome.explore.cazome_sizes.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest

from argparse import Namespace

from cazomevolve.cazome.explore import cazy_families


def test_build_fam_freq_df(fam_freq_df_with_tax):
    assert len(cazy_families.build_fam_freq_df(fam_freq_df_with_tax, ['Genus', 'Species'])) == 1


def test_build_fam_fbuild_row_coloursreq_df(built_fam_freq_df):
    cazy_families.build_row_colours(built_fam_freq_df, 'Genus', 'Set1')


def test_build_clustermap(built_fam_freq_df):
    row_colours, lut = cazy_families.build_row_colours(built_fam_freq_df, 'Genus', 'Set1')
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    cazy_families.build_family_clustermap(
        built_fam_freq_df,
        row_colours=row_colours,
        fig_size=(10,10),
        lut=lut,
    )


def test_build_family_clustermap_multi_legend(built_fam_freq_df):
    row_colours, lut = cazy_families.build_row_colours(built_fam_freq_df, 'Genus', 'Set1')
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome', 'Species'])
    cazy_families.build_family_clustermap_multi_legend(
        built_fam_freq_df,
        row_colours=[row_colours],
        fig_size=(10,10),
        luts=[lut],
        legend_titles=['test1', 'test2'],
        bbox_to_anchors=[(0,0), (1,1)],
    )


def test_core_cazome(built_fam_freq_df):
    assert len(cazy_families.identify_core_cazome(built_fam_freq_df)) == 20


def test_boxplot(built_fam_freq_df):
    df = built_fam_freq_df[['GH1', 'PL1']]
    cazy_families.plot_fam_boxplot(df)


def test_build_fam_mean_freq_df(built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.drop('Species', axis=1)
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome'])
    cazy_families.build_fam_mean_freq_df(built_fam_freq_df, 'Genus', round_by=2)


def test_get_group_specific_fams(built_fam_freq_df):
    built_fam_freq_df = built_fam_freq_df.drop('Species', axis=1)
    built_fam_freq_df = built_fam_freq_df.set_index(['Genome'])
    fams = list(built_fam_freq_df.columns)
    cazy_families.get_group_specific_fams(built_fam_freq_df, 'Genus', fams)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Test taxs module ncbi.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import pandas as pd

from argparse import Namespace

from saintBioutils.utilities import logger

from cazomevolve.taxs import ncbi


@pytest.fixture
def argsdict():
    return {'args': Namespace(
        retries=2,
    )}


@pytest.fixture
def col_names():
    return ['Genome', 'Kingdom', 'Genus', 'Species']

@pytest.fixture
def col_names_full():
    return ['Genome', 'Kingdom', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']


def test_add_ncbi(col_names, monkeypatch):
    def mock_get(*args, **kwards):
        return {}, ['genome']
    monkeypatch.setattr(ncbi, "get_tax_ids", mock_get)
    monkeypatch.setattr(ncbi, "get_ncbi_taxs", mock_get)

    assert {'genome': 'genome_NaN_NaN_NaN'} == ncbi.add_ncbi_taxs({}, [], col_names, 'args')
    

def test_get_ids_failed(argsdict, monkeypatch):
    """Test getting tax ids when connection fails"""
    def mock_entrez_tax_call(*args, **kwargs):
        """Mocks call to Entrez."""
        return
    
    monkeypatch.setattr(ncbi, "entrez_retry", mock_entrez_tax_call)

    out1, out2 = ncbi.get_tax_ids({'genomes'}, argsdict['args'])
    assert out1 == {}
    assert out2 == ['genomes']


def test_get_ncbi_taxs_failed(argsdict, col_names_full, monkeypatch):
    """Test getting taxs when connection fails"""
    def mock_entrez_tax_call(*args, **kwargs):
        """Mocks call to Entrez."""
        return
    
    monkeypatch.setattr(ncbi, "entrez_retry", mock_entrez_tax_call)

    out1, out2 = ncbi.get_ncbi_taxs({}, {}, {'genomes'}, col_names_full, argsdict['args'])
    assert out1 == {}
    assert out2 == {'genomes'}


def test_get_ncbi_taxs(argsdict, col_names_full, monkeypatch, test_input_dir):
    """Test getting taxs when connection fails"""
    ncbi_result = test_input_dir / "ncbi/ncbi_record.xml"

    with open(ncbi_result, "rb") as fh:
        result = fh
        def mock_entrez(*args, **kwards):
            return result
    
        monkeypatch.setattr(ncbi, "entrez_retry", mock_entrez)

        out1, out2 = ncbi.get_ncbi_taxs({}, {}, {'genomes'}, col_names_full, argsdict['args'])
        assert out1 == {}
        assert out2 == {'genomes'}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Test downloading genomes using download_genomes.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import pandas as pd
import logging

from argparse import Namespace

from saintBioutils.utilities import logger

from cazomevolve.genomes import download_genomes


@pytest.fixture
def argsdict(test_output_dir):
    return {'args': Namespace(
        email='dummy',
        retries=2,
        output_dir=test_output_dir,
        force=False,
        nodelete=True,
        terms='Aspergillus,Trichoderma'
    )}


@pytest.fixture
def col_names():
    return ['Genome', 'Kingdom', 'Genus', 'Species']

@pytest.fixture
def col_names_full():
    return ['Genome', 'Kingdom', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']



def test_download_genomes_main(argsdict, monkeypatch):
    def mock_none(*args, **kwards):
        return
    def mock_get_ids(*args, **kwards):
        return [1]

    monkeypatch.setattr(download_genomes, "make_output_directory", mock_none)
    monkeypatch.setattr(download_genomes, "get_id_list", mock_get_ids)
    monkeypatch.setattr(download_genomes, "get_tax_ids", mock_none)
    monkeypatch.setattr(download_genomes, "closing_message", mock_none)

    download_genomes.main(args=argsdict['args'])


def test_get_ids(monkeypatch, test_input_dir):
    """Test getting taxs when connection fails"""
    ncbi_result = test_input_dir / "ncbi/ncbi_esearch.xml"

    with open(ncbi_result, "rb") as fh:
        result = fh
        def mock_entrez(*args, **kwards):
            return result
    
        monkeypatch.setattr(download_genomes, "entrez_retry", mock_entrez)

        out = download_genomes.get_id_list('Aspergillus')
        out.sort()
        target = ['17020241', '16863991', '16863981', '16863971', '16863961', '16863951', '16599221', '16585761', '16567171', '16567101']
        target.sort()
        assert target == out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Run all-vs-all seq comparison using DIAMOND"""


import argparse
import subprocess

from inspect import getsourcefile
from os.path import abspath


def main(args: argparse.Namespace) -> int:
    cazevolve_path = abspath(getsourcefile(lambda:0).replace("scripts/run_fam_diamond.py","seq_diversity/run_diamond.sh"))

    cmd = [
            cazevolve_path,
            args.fasta,
            args.diamond_db,
            args.outfile,
        ]
    print(f"Running command: {' '.join(cmd)}")

    theproc = subprocess.call(args=cmd)
    
    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, _SubParsersAction
from pathlib import Path
from typing import List, Optional

from cazomevolve.taxs import add_taxs


def build_parser(
    subps: _SubParsersAction, parents: Optional[List[ArgumentParser]] = None
) -> None:
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = subps.add_parser(
        "add_taxs", formatter_class=ArgumentDefaultsHelpFormatter
    )

    # Add positional arguments to parser
    parser.add_argument(
        "email",
        type=str,
        help="User email address (Required by NCBI)",
    )

    parser.add_argument(
        "--FG_FILE",
        type=Path,
        help="Path to tab delim list of family-genome",
    )

    parser.add_argument(
        "--FGP_FILE",
        type=Path,
        help="Path to tab delim list of family-genome-protein",
    )

    parser.add_argument(
        "--gtdb",
        type=Path,
        help="Path to gtdb database download TSV file (from https://data.gtdb.ecogenomic.org/)",
    )

    parser.add_argument(
        "--outpath",
        type=Path,
        help=(
            "Path to write out CSV of retrieved tax data.\n"
            "Else, writes out to taxomines.csv\n"
            "in the same dir as --FGP_FILE/--FG_FILE"
        ),
    )

    parser.add_argument(
        "--kingdom",
        dest="kingdom",
        action="store_true",
        default=False,
        help="Retrieve and add Kingdom lineage",
    )

    parser.add_argument(
        "--phylum",
        dest="phylum",
        action="store_true",
        default=False,
        help="Retrieve and add phylum lineage",
    )

    parser.add_argument(
        "--tax_class",
        dest="tax_class",
        action="store_true",
        default=False,
        help="Retrieve and add class lineage",
    )

    parser.add_argument(
        "--tax_order",
        dest="tax_order",
        action="store_true",
        default=False,
        help="Retrieve and add order lineage",
    )

    parser.add_argument(
        "--tax_family",
        dest="tax_family",
        action="store_true",
        default=False,
        help="Retrieve and add (tax) family lineage",
    )

    parser.add_argument(
        "--genus",
        dest="genus",
        action="store_true",
        default=False,
        help="Retrieve and add genus lineage",
    )

    parser.add_argument(
        "--species",
        dest="species",
        action="store_true",
        default=False,
        help="Retrieve and add species lineage",
    )

    # Add option to specific directory for log to be written out to
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        metavar="log file name",
        default=None,
        help="Defines log file name and/or path",
    )

    parser.add_argument(
        "--retries",
        dest="retries",
        type=int,
        default=10,
        help="Number of times to retry a failed connection to NCBI",
    )

    # Add option to specify verbose logging
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Set logger level to 'INFO'",
    )

    parser.set_defaults(func=add_taxs.main)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests seq diversty module

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import pandas as pd

from argparse import Namespace
from bs4 import BeautifulSoup
from requests.exceptions import MissingSchema

from cazomevolve.seq_diversity.explore import cazy, parse, plot


#
# CAZY
#


def test_get_proteins(test_input_dir):
    fasta = test_input_dir / "seq_diversity/fasta.faa"
    out = cazy.get_cazy_proteins(fasta)
    out.sort()
    expected = ['CAG72928.1', 'CAG72925.1', 'CAG72926.1', 'CAG72927.1', 'CAG72929.1']
    expected.sort()
    assert out == expected


def test_get_db_prots_fail(monkeypatch):
    """Test get_cazy_db_prots when connection fails"""
    def mock_get_page(*args, **kwards):
        return None, 'ERROR'

    monkeypatch.setattr(cazy, "get_page", mock_get_page)

    assert [] == cazy.get_cazy_db_prots('PL1', characterised=True, structured=True)


def test_get_db_prots(test_input_dir, monkeypatch):
    """Test get_cazy_db_prots when connection fails"""
    _path = test_input_dir / "seq_diversity/CAZy-PL1.html"

    with open(_path) as fp:
        soup = BeautifulSoup(fp, features="lxml")

    def mock_get_page(*args, **kwards):
        return soup, None

    monkeypatch.setattr(cazy, "get_page", mock_get_page)

    assert len(cazy.get_cazy_db_prots('PL1', characterised=True)) == 419


def test_browser_decorator():
    """Test browser_decorator to ensure proper handling if unsuccessful."""
    args = {"args": Namespace(timeout=2)}
    result = cazy.get_page('www.caz!!!!!!!!y.org', max_tries=1)
    assert True == (result[0] is None) and (type(result[1]) is MissingSchema)


#
# PARSE
#


def test_load_data(test_input_dir, monkeypatch):
    _path = test_input_dir / "seq_diversity/diamond_output"
    
    parse.load_data(_path, "PL4")


def test_remove_redundant(test_input_dir):
    _path = test_input_dir / "seq_diversity/diamond_output"

    df = parse.load_data(_path, "PL4")
    parse.remove_redunant_prots(df, 'PL4', structured_prots={'PL4': 'QHQ22566.1'}, candidates={'PL4': 'QHQ22566.1'}, characterised_prots={'PL4': 'BAC69819.1'})

#
# PLOT
#


def test_clustermap(test_input_dir):
    _path = test_input_dir / "seq_diversity/parsed_data.csv"
    df = pd.read_csv(_path, index_col="Unnamed: 0")

    fig = plot.plot_clustermap(
        df=df,
        fam='PL4',
        varaible='BSR',
        title='Test',
        annotate=True,
        char_only=True,
        structured_prots={'PL4': ['QHQ22566.1']},
        candidates={'PL4': ['QHQ22566.1', 'ACD60718.1', 'BAE63118.1', 'AXQ10473.1', 'ATS40189.1', 'CAG7898885.1', 'UWI46514.1']},
        characterised_prots={'PL4': ['BAC69819.1']},
    )

    plot.plot_heatmap_of_clustermap(
        fig=fig,
        df=df,
        fam='PL4',
        varaible='BSR',
        title='Test',
        annotate=True,
        char_only=True,
        structured_prots={'PL4': ['QHQ22566.1']},
        candidates={'PL4': ['QHQ22566.1', 'ACD60718.1', 'BAE63118.1', 'AXQ10473.1', 'ATS40189.1', 'CAG7898885.1', 'UWI46514.1']},
        characterised_prots={'PL4': ['BAC69819.1']},
    )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions for calculating CAZome sizes and proteome proportions"""


import pandas as pd
import numpy as np
import re

from Bio import SeqIO
from saintBioutils.utilities.file_io.get_paths import get_file_paths
from tqdm import tqdm


def count_items_in_cazome(gfp_df, item, grp, round_by=None):
    """Count the number of unique items per genome and per specificed tax grouping
    
    :param gfp_df: panda df, cols = ['Family', 'Genome', 'Protein', 'tax grp', 'tax grp'...]
    :param item: str, name of column to calculate incidence for, e.g. 'Protein' or 'Family'
    :param grp: str, name of column to group genomes by
    :param round_by: int, number of figures to round mean and sd by. If None do not round
    
    Return
    * dict of {grp: {genome: {'items': {items}, 'numOfItems': int(num of items)}}}
    * df, cols = []
    """
    cazome_sizes = {}  # {genus: {genome: {'proteins': unique prot acc, 'numOfcazymes': int(num of prots)}}}

    for ri in tqdm(range(len(gfp_df)), desc="Gathering CAZy families per genome"):
        row = gfp_df.iloc[ri]
        genome = row['Genome']
        row_item = row[item]
        grp_name = row[grp]

        try:
            cazome_sizes[grp_name]
        except KeyError:
            cazome_sizes[grp_name] = {}

        try:
            cazome_sizes[grp_name][genome][f'{item}s'].add(row_item)
        except KeyError:
            cazome_sizes[grp_name][genome] = {f"{item}s": {row_item}}
            
    cazyme_size_data = []

    for grp_name in tqdm(cazome_sizes, desc=f"Calculating num of {item} per genome and per {grp}"):
        for genome in cazome_sizes[grp_name]:
            num_of_items = len(cazome_sizes[grp_name][genome][f'{item}s'])
            cazome_sizes[grp_name][genome][f'numOf{item}s'] = num_of_items
            
        num_of_items = []
        for genome in cazome_sizes[grp_name]:
            num_of_items.append(cazome_sizes[grp_name][genome][f'numOf{item}s'])

        mean_items = np.mean(num_of_items)
        sd_items = np.std(num_of_items)
        
        if round_by is not None:
            mean_items = mean_items.round(decimals = round_by)
            sd_items = sd_items.round(decimals = round_by)
        
        num_genomes = len(list(cazome_sizes[grp_name].keys()))
        
        new_row = [grp_name, mean_items, sd_items, num_genomes]
        cazyme_size_data.append(new_row)
    
    cols = [grp, f'Mean{item}s', f'Sd{item}s', 'NumOfGenomes']
    cazome_size_df = pd.DataFrame(cazyme_size_data, columns=cols)
    
    return cazome_sizes, cazome_size_df


def get_proteome_sizes(proteome_dir, gfp_df, grp):
    """Count the number of proteins in the proteome file of each genome
    
    Build a dict of proteome sizes grouped by tax lineage ('grp')
    
    :param proteome_dir: Path or str, path to dir containing .faa proteome files
    :param gfp_df: pandas df containing families, genomes, tax_rank, tank_rank...
    :param grp: str, name of column (tax_rank) to group genomes by, e.g. 'Genus'
    
    Return dict {grp: {genome: {'numOfproteins': int()}}}"""
    proteome_files = get_file_paths(proteome_dir, suffixes=['.faa'])

    proteome_sizes = {}  # {grp: {genome: {'numOfproteins': int()}}}

    for fasta_path in tqdm(proteome_files, desc="Getting proteome sizes"):
        try:
            genome = re.findall(r"GCF_\d+\.\d{1,5}", fasta_path.name)[0]
        except IndexError:
            try:
                genome = re.findall(r"GCA_\d+\.\d{1,5}", fasta_path.name)[0]
            except IndexError:
                print(
                    f"Could not retrieve genomic accession from\n{fasta_path}\n"
                    "Skipping FASTA file"
                )
                continue

        # get tax group
        tax_row = gfp_df[gfp_df['Genome'] == genome]
        if len(tax_row) == 0:
            print(
                f"Genome {genome} was not in the FGP file."
                "Therefore, skipping proteome proporiton count for genome"
            )
            continue
        tax_group = tax_row[grp].values[0]

        num_of_proteins = 0
        for record in SeqIO.parse(fasta_path, 'fasta'):
            num_of_proteins += 1

        try:
            proteome_sizes[tax_group]
        except KeyError:
            proteome_sizes[tax_group] = {}

        proteome_sizes[tax_group][genome] = {'numOfProteins': num_of_proteins}

    return proteome_sizes


def calc_proteome_representation(proteome_dict, cazome_sizes_dict, grp, round_by=None):
    """Calculate the percentage of the proteome represented by the CAZome per genome
    and the mean per tax group ('grp')
    
    :param proteome_dict: dict {grp: {genome: {'numOfproteins': int()}}}
    :param cazome_sizes_dict: dict {grp: {genome: {'Proteins': {prots}, 'numOfProteins': int(num of prots)}}}
    :param grp: str, name of column (tax_rank) to group genomes by, e.g. 'Genus'
    :param round_by: int, num of dp to round means and sd to, if none does not round
    
    Return
    * df with columns [grp, 'MeanProteomeSize', 'SdProteomeSize', 'MeanProteomePerc', 'SdProteomePerc', 'NumOfGenomes']
    """
    proteome_perc_dict = {}  # {grp_name: {genome: {'numOfProteins': int(num prot ids)}}}

    for grp_name in tqdm(cazome_sizes_dict, desc='Getting proteome size'):
        for genome in cazome_sizes_dict[grp_name]:
            # gather num of proteins in CAZome and in the proteome
            cazome_size = cazome_sizes_dict[grp_name][genome]['numOfProteins']
            try:
                proteome_size = proteome_dict[grp_name][genome]['numOfProteins']
            except KeyError:
                print(f"Not proteome size available for {genome}.\nSkipping this genome")
                continue

            percentage = (cazome_size / proteome_size) * 100
            
            # check if grp name is in dict
            try:
                proteome_perc_dict[grp_name]
            except KeyError:
                proteome_perc_dict[grp_name] = {}
            
            # add percentage of the proteome to the dict
            # round only when calc mean
            proteome_perc_dict[grp_name][genome] = percentage
    
    # calculate the mean proteome size and percentage per tax group (grp)
    # cazome_sizes_dict = {grp: {genome: {'Proteins': {prots}, 'numOfProteins': int(num of prots)}}}

    df_data = []  # [[grp_name, mean proteome size, sd proteome size, mean perc, sd perc]]

    for grp_name in tqdm(proteome_perc_dict, desc='Calc proteome perc'):
        # gather proteome sizes and perc for the grp
        grp_proteome_size = []
        grp_proteome_perc = []
        
        for genome in proteome_perc_dict[grp_name]:
            grp_proteome_size.append(proteome_dict[grp_name][genome]['numOfProteins'])
            grp_proteome_perc.append(proteome_perc_dict[grp_name][genome])

        # calc means and sd
        mean_proteome = np.mean(grp_proteome_size)
        sd_proteome = np.std(grp_proteome_size)

        mean_perc = np.mean(grp_proteome_perc)
        sd_perc = np.std(grp_proteome_perc)

        if round_by is not None:
            mean_proteome = round(mean_proteome, round_by)
            sd_proteome = round(sd_proteome, round_by)
            mean_perc = round(mean_perc, round_by)
            sd_perc = round(sd_perc, round_by)

        num_of_genomes = len(list(proteome_perc_dict[grp_name].keys()))

        df_data.append(
            [grp_name, mean_proteome, sd_proteome, mean_perc, sd_perc, num_of_genomes]
        )

    col_names = [grp, 'MeanProteomeSize', 'SdProteomeSize', 'MeanProteomePerc', 'SdProteomePerc', 'NumOfGenomes']
    df = pd.DataFrame(df_data, columns=col_names)
    
    return df


def count_cazyme_fam_ratio(fgp_df, grp, round_by=None):
    """Calculate the mean (and SD) CAZyme to CAZy family ratio across the genomes for each group e.g. genus
    
    :param fgp_df: panda df, cols = ['Family', 'Genome', 'Protein', 'tax grp', 'tax grp'...]
    :param grp: str, name of column to group genomes by
    :param round_by: int, number of figures to round mean and sd by. If None do not round
    
    Return
    * dict of {grp: {genome: {'items': {items}, 'numOfItems': int(num of items)}}}
    * df, cols = []
    """
    cazome_sizes = {}  # {genus: {genome: {'proteins': unique prot acc, 'numOfcazymes': int(num of prots)}}}
    
    for ri in tqdm(range(len(fgp_df)), desc="Gathering CAZymes and CAZy families per genome"):
        row = fgp_df.iloc[ri]
        genome = row['Genome']
        fam = row['Family']
        protein = row['Protein']
        grp_name = row[grp]

        try:
            cazome_sizes[grp_name]
        except KeyError:
            cazome_sizes[grp_name] = {}

        try:
            cazome_sizes[grp_name][genome]['Proteins'].add(protein)
            cazome_sizes[grp_name][genome]['Families'].add(fam)
        except KeyError:
            cazome_sizes[grp_name][genome] = {'Proteins': {protein}, 'Families': {fam}}
            
    cazyme_ratio_data = []

    for grp_name in tqdm(cazome_sizes, desc=f"Calculating CAZyme/CAZy family ratio"):
        for genome in cazome_sizes[grp_name]:
            num_of_cazymes = len(cazome_sizes[grp_name][genome]['Proteins'])
            num_of_families = len(cazome_sizes[grp_name][genome]['Families'])
            cazyme_fam_ratio = num_of_cazymes / num_of_families
            cazome_sizes[grp_name][genome]['ratio'] =cazyme_fam_ratio
            
        ratios = []
        for genome in cazome_sizes[grp_name]:
            ratios.append(cazome_sizes[grp_name][genome]['ratio'])

        mean_ratio = np.mean(ratios)
        sd_ratio = np.std(ratios)
        
        if round_by is not None:
            mean_ratio = mean_ratio.round(decimals = round_by)
            sd_ratio = sd_ratio.round(decimals = round_by)
        
        num_genomes = len(list(cazome_sizes[grp_name].keys()))
        
        new_row = [grp_name, mean_ratio, sd_ratio, num_genomes]
        cazyme_ratio_data.append(new_row)
    
    cols = [grp, 'MeanCAZymeToFamRatio', 'SdCAZymeToFamRatio', 'NumOfGenomes']
    cazome_ratio_df = pd.DataFrame(cazyme_ratio_data, columns=cols)
    
    return cazome_sizes, cazome_ratio_df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Script for invoking dbCAN"""


import re
import subprocess
import logging

from tqdm import tqdm
from typing import List, Optional

from saintBioutils.utilities.file_io import make_output_directory
from saintBioutils.utilities.file_io.get_paths import get_file_paths

from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    make_output_directory(args.output_dir, args.force, args.nodelete)

    # get the path to every FASTA to be parsed by dbCAN
    fasta_files_paths = list(set(get_file_paths(args.input_dir, suffixes=['fasta', 'faa'])))
    fasta_files_paths.sort()
    print(f"Retrieved {len(fasta_files_paths)} fasta files from {args.input_dir}")

    for fasta_path in tqdm(fasta_files_paths, desc="Running dbCAN"):
        # define path to output dir that will house output for this specific input FASTA file
        # extract genomic accession from the file name, and name output dir after the accession
        try:
            genomic_accession = re.findall(r"GCF_\d+\.\d{1,5}", fasta_path.name)[0]
        except IndexError:
            try:
                genomic_accession = re.findall(r"GCA_\d+\.\d{1,5}", fasta_path.name)[0]
            except IndexError:
                logger.warning(
                    f"Could not retrieve genomic accession from\n{fasta_path}\n"
                    "Skipping FASTA file"
                )
                continue
        
        output_dir = args.output_dir / genomic_accession

        if output_dir.exists():
            print(f"Already parsed {genomic_accession}\nSKIIIP")
            continue

        invoke_dbcan(fasta_path, output_dir, args)

    closing_message('Invoke dbCAN', args)


def invoke_dbcan(input_path, out_dir, args):
    """Invoke the prediction tool (run-)dbCAN.

    :param input_path: path to input FASTA file
    :param out_dir: path to output directory for input FASTA file query
    :param args: cmd-line args parser

    Return nothing
    """
    # make the output directory
    make_output_directory(out_dir, True, False)

    if args.dbcan_version == 2:
        # create list of args to invoke run_dbCAN
        dbcan_args = [
            "run_dbcan.py",
            str(input_path),
            "protein",
            "--out_dir",
            str(out_dir),
        ]

    elif args.dbcan_version == 3:
        # create list of args to invoke run_dbCAN
        dbcan_args = [
            "run_dbcan",
            str(input_path),
            "protein",
            "--out_dir",
            str(out_dir),
            "--stp_cpu",
            str(args.cpu),
            "--tf_cpu",
            str(args.cpu),
            "--eCAMI_jobs",
            str(args.cpu),
            "--hmm_cpu",
            str(args.cpu),
            "--dia_cpu",
            str(args.cpu),
        ]

    else:
        # create list of args to invoke run_dbCAN
        dbcan_args = [
            "run_dbcan",
            str(input_path),
            "protein",
            "--out_dir",
            str(out_dir),
            "--stp_cpu",
            str(args.cpu),
            "--tf_cpu",
            str(args.cpu),
            "--hmm_cpu",
            str(args.cpu),
            "--dia_cpu",
            str(args.cpu),
        ]

    with open(f"{out_dir}/dbcan.log", "w+") as fh:
        process = subprocess.run(dbcan_args, text=True, capture_output=True)

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Build CLI for invoke_dbcan.py"""


import multiprocessing

from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, _SubParsersAction
from pathlib import Path
from typing import List, Optional

from cazomevolve.cazome.dbcan import invoke_dbcan


def build_parser(
    subps: _SubParsersAction, parents: Optional[List[ArgumentParser]] = None
) -> None:
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = subps.add_parser(
        "run_dbcan", formatter_class=ArgumentDefaultsHelpFormatter
    )

    # Add positional arguments to parser

    parser.add_argument(
        "input_dir",
        type=Path,
        help="Path to directory containing FASTAs to be parsed by dbCAN",
    )

    parser.add_argument(
        "output_dir",
        type=Path,
        help="Path to directory to write out genomic assemblies",
    )

    parser.add_argument(
        "dbcan_version",
        type=int,
        help="Major version number of dbCAN being used (e.g. 2, 3, or 4)",
    )

    parser.add_argument(
        "--cpu",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Number of CPU cores to use. Default all available cores",
    )

    # Add option to force file over writting
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force file over writting",
    )

    # Add option to specific directory for log to be written out to
    parser.add_argument(
        "-l",
        "--log",
        type=Path,
        metavar="log file name",
        default=None,
        help="Defines log file name and/or path",
    )
    # Add option to prevent over writing of existing files
    # and cause addition of files to output directory
    parser.add_argument(
        "-n",
        "--nodelete",
        dest="nodelete",
        action="store_true",
        default=False,
        help="enable/disable deletion of exisiting files",
    )

    # Add option to specify verbose logging
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Set logger level to 'INFO'",
    )
    parser.set_defaults(func=invoke_dbcan.main)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Query NCBI to retrieve taxonomic classifications for Assembly records"""


import logging

from Bio import Entrez

from saintBioutils.genbank import entrez_retry
from tqdm import tqdm


def add_ncbi_taxs(genomes_tax_dict, genomes_to_query, col_names, args):
    """Query NCBI to get the taxonomic classification and add to {genome: f"{genome}_{tax}"}

    :param genomes_tax_dict: dict, {genome: f"{genome}_{tax}"}  - genomes with tax classification in gtdb
    :param genomes_to_query: set of genomic acc to query ncbi with to get tax classification
    :param col_names: list of lineage ranks
    :param args: cli args parser

    Return genomes_tax_dict
    """
    taxids_genomes, failed_genomes = get_tax_ids(genomes_to_query, args)

    genomes_tax_dict, failed_genomes = get_ncbi_taxs(taxids_genomes, genomes_tax_dict, failed_genomes, col_names, args)

    for genome in failed_genomes:
        genomes_tax_dict[genome] = f"{genome}_"
        for name in col_names[1:]:
            genomes_tax_dict[genome] += f"NaN_"
        genomes_tax_dict[genome] = genomes_tax_dict[genome][:-1]

    return genomes_tax_dict


def get_tax_ids(genomes, args):
    """Get NCBI Tax IDs for  genomes.

    :param genomes: list of genomic assembly accessions
    :param args: cli args parser

    Return dict of {tax id: {genomes}} and list of genomes for which tax records could not be retrieved
    """
    logger = logging.getLogger(__name__)
    taxids_genomes = {}  # {tax id: {genomes}}
    failed_genomes = []

    for genome in tqdm(genomes, desc="Getting tax ids"):
        # retrieve the ID of corresponding record in NCBI Assembly
        try:
            with entrez_retry(
                args.retries,
                Entrez.esearch,
                db="Assembly",
                term=genome,
            ) as accession_handle:
                record_meta_data = Entrez.read(accession_handle, validate=False)
        except (TypeError, AttributeError) as error:
            logger.warning(f"Could not retrieve tax data for {genome}")
            failed_genomes.append(genome)
            continue

        genome_record_id = record_meta_data['IdList'][0]

        # Fetch the record from the Assembly db, by querying by the record ID
        try:
            with entrez_retry(
                args.retries,
                Entrez.efetch,
                db="Assembly",
                id=genome_record_id,
                rettype="docsum",
            ) as accession_handle:
                accession_record = Entrez.read(accession_handle, validate=False)
        except (TypeError, AttributeError) as error:
            logger.warning(f"Could not fetch tax data for {genome}\nError:{error}")
            failed_genomes.append(genome)
            continue
        
        taxid = accession_record['DocumentSummarySet']['DocumentSummary'][0]['Taxid']
        
        try:
            taxids_genomes[taxid].add(genome)
        except KeyError:
            taxids_genomes[taxid] = {genome}

    return taxids_genomes, failed_genomes


def get_ncbi_taxs(taxids_genomes, genomes_tax_dict, failed_genomes, col_names, args):
    """Retrieve lineage data from NCBI Taxonomy db

    :param taxid_genomes: dict {taxid: {genomes}}
    :param genomes_tax_dict: dict, {genome: f"{genome}_{tax}"}  - genomes with tax classification in gtdb
    :param failed_genomes: list of genomes for which tax data could not be retrieved from NCBI
    :param col_names: list of lineage ranks to retrieve
    :param args: cli args parser

    Return genomes_tax_dict {genome: f"{genome}_{tax}"}
    """
    logger = logging.getLogger(__name__)

    ranks = []
    if 'Kingdom' in col_names:
        ranks.append('superkingdom')
    if 'Phylum' in col_names:
        ranks.append('phylum')
    if 'Class' in col_names:
        ranks.append('class')
    if 'Order' in col_names:
        ranks.append('order')
    if 'Family' in col_names:
        ranks.append('family')
    if 'Genus' in col_names:
        ranks.append('genus')
    # retrieve species from scientific name (minus genus)

    for taxid in tqdm(taxids_genomes, desc="Getting taxonomies"):
        try:
            with entrez_retry(
                args.retries,
                Entrez.efetch,
                db="Taxonomy",
                id=taxid,
                # rettype="docsum",
            ) as handle:
                tax_record = Entrez.read(handle, validate=False)
        except (TypeError, AttributeError) as error:
            logger.warning(f"Could not fetch tax data for {genome}\nError:{error}")
            for genome in taxids_genomes[taxid]:
                failed_genomes.append(genome)
            continue
        tax_info = {}
        for rank in ranks:
            added = False
            for feature in tax_record[0]['LineageEx']:
                if feature['Rank'] == rank:
                    tax_info[rank] = feature['ScientificName']
                    added = True
                    break
            if added is False:
                tax_info[rank] = 'NaN'

        for genome in taxids_genomes[taxid]:
            genome_tax = f"{genome}_"
            for rank in tax_info:
                genome_tax += f"{tax_info[rank]}_"

            if 'Species' in col_names:
                scientific_name = tax_record[0]['ScientificName']
                genome_tax += f'{" ".join(scientific_name.split(" ")[1:])}_'

            genome_tax = genome_tax[:-1]

            genomes_tax_dict[genome] = genome_tax

    return genomes_tax_dict, failed_genomes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Retrieve all genomic assembly accessions descendent from a taxonomy node"""


import argparse
import sys
import subprocess
import os

from inspect import getsourcefile
from os.path import abspath
from pathlib import Path
from typing import List, Optional

from saintBioutils.utilities.file_io import make_output_directory


# $1 path to file listing the accessions, with a unique genome accession per row
# $2 output directory to write out the genomes to
# $3 file options - A comma-separated list of formats is also possible. For example: "fasta,assembly-report". Choose from:
# ['genbank', 'fasta', 'rm', 'features', 'gff', 'protein-fasta', 'genpept', 'wgs', 'cds-fasta', 'rna-fna', 'rna-fasta', 'assembly-report', 'assembly-stats', 'all']
# $4 refseq or genbank
# $5 assembly level, default all, ['all', 'complete', 'chromosome', 'scaffold', 'contig']


def main(args: argparse.Namespace) -> int:
    if str(Path(args.outdir).parent) != ".":
        make_output_directory(Path(args.outdir), args.force, args.nodelete)

    cazevolve_path = abspath(getsourcefile(lambda:0).replace("scripts/download_acc_genomes.py","scripts/bash/download_acc_genomes.sh"))

    cmd = [
            cazevolve_path,
            args.accessions,
            args.outdir,
            args.file_opts,
            args.database,
            args.assembly_levels,
        ]

    print(f"Running command: {' '.join(['download_acc_genomes.sh']+cmd[1:])}")

    theproc = subprocess.call(
        [
            cazevolve_path,
            args.accessions,
            args.outdir,
            args.file_opts,
            args.database,
            args.assembly_levels,
        ]
    )  

    return 0

if __name__ == "__main__":
    main()



# /usr/bin/env python
# -*- coding: utf-8 -*-

# (c) University of St Andrews 2023
# (c) University of Strathclyde 2023
# (c) James Hutton Institute 2023
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""extract_cds.py

Script to match aligned single-copy orthologue protein sequences from
MAFFT alignments with the corresponding CDS sequences, prior to
backtranslation/threading with t-coffee.
"""
import os

from pathlib import Path

from Bio import SeqIO
from tqdm import tqdm

from saintBioutils.utilities.file_io import make_output_directory

import argparse

from pathlib import Path
from typing import List, Optional



def main(argv: Optional[List[str]] = None):
    if argv is None:
        parser = build_parser()
        args = parser.parse_args()
    else:
        parser = build_parser(argv)
        args = parser.parse_args()

    make_output_directory(args.OUTDIR, force=args.force, nodelete=args.nodelete)

    # args.PROTDIR contains the MAFFT alignments
    args.PROTDIR = Path("orthologues/Results_July15/Single_Copy_Orthologue_Sequences")

    # args.CDSDIR contains the CDS/proteome for each genome
    args.CDSDIR = Path("data/pecto_dic/tree/genomes/cds")

    # args.OUTDIR is the location to which CDS sequences corresponding to the
    # protein alignments will be written
    args.OUTDIR = Path("data/pecto_dic/tree/sco_cds")

    # The protein sequences have IDs in the form NC_004547.2_1; this is not
    # enough to identify the corresponding genome sequence/CDS file. We take
    # a brute force approach and load all CDS into memory
    print("Loading CDS sequences into memory.")
    seqdict = {}
    for cdsfpath in tqdm(args.CDSDIR.iterdir()):
        seqdict.update(SeqIO.to_dict(SeqIO.parse(cdsfpath, "fasta")))
    print(f"Loaded {len(seqdict)} CDS sequences.")

    # We iterate over the SCO protein files, and extract the corresponding
    # CDS sequences for each, then write these to args.OUTDIR
    print(f"Writing extracted CDS files to {args.OUTDIR}")
    os.makedirs(args.OUTDIR, exist_ok=True)  # Create output directory, if needed
    for protfpath in tqdm(args.PROTDIR.iterdir()):
        with (args.OUTDIR / protfpath.stem).with_suffix(".fasta").open("w") as ofh:
            cds = [seqdict[_.id] for _ in SeqIO.parse(protfpath, "fasta")]
            fixed_cds = []
            for _ in cds:
                if str(_.seq[-3:]) not in ("TAA", "TAG", "TGA"):
                    fixed_cds.append(_)
                else:
                    fixed_cds.append(_[:-3])
            SeqIO.write(
                fixed_cds,
                ofh,
                "fasta",
            )



def build_parser(argv: Optional[List] = None):
    """Return ArgumentParser parser for script."""
    # Create parser object
    parser = argparse.ArgumentParser(
        prog="cw_query_database",
        description="Interrogate a local CAZyme database",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "PROTDIR",
        type=Path,
        help="Path to dir containing SCO protein seqs"
    )
    parser.add_argument(
        "CDSDIR",
        type=Path,
        help="Path to dir containing prodigal predicted CDS seqs"
    )
    parser.add_argument(
        "OUTDIR",
        type=Path,
        help="Path to dir to write out selecte CDS seqs"
    )

    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force writing to existing output dir",
    )

    parser.add_argument(
        "-n",
        "--nodelete",
        dest="nodelete",
        action="store_true",
        default=False,
        help="When called, content in the existing output dir is NOT deleted",
    )

    if argv is None:
        # parse command-line
        return parser
    else:
        # return namespace
        return parser.parse_args(argv)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scazome.explore.cazome_sizes.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest

from argparse import Namespace

from cazomevolve.cazome.explore import cazome_sizes


def test_count_items(fam_freq_df_with_tax):
    size_dict, var = cazome_sizes.count_items_in_cazome(fam_freq_df_with_tax, 'Protein', 'Genus', round_by=2)
    assert len(size_dict) == 1


def test_get_proteome_sizes(test_input_dir, fam_freq_df_with_tax):
    _path = test_input_dir / "cazome_explore"

    assert len(cazome_sizes.get_proteome_sizes(_path, fam_freq_df_with_tax, 'Genus')) == 1


def test_calc_proteome_represent(fam_freq_df_with_tax, test_input_dir):
    _path = test_input_dir / "cazome_explore"
    size_dict, var = cazome_sizes.count_items_in_cazome(fam_freq_df_with_tax, 'Protein', 'Genus', round_by=2)
    prot_dict = cazome_sizes.get_proteome_sizes(_path, fam_freq_df_with_tax, 'Genus')

    cazome_sizes.calc_proteome_representation(prot_dict, size_dict, 'Genus', 2)


def test_fam_ratio(fam_freq_df_with_tax):
    cazome_sizes.count_cazyme_fam_ratio(fam_freq_df_with_tax, 'Genus', 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions for parsing output from cazomevolve and generating alternative structure architectures"""


import pandas as pd

from copy import copy
from collections import Counter

from tqdm import tqdm
from saintBioutils.utilities.file_io.get_paths import get_dir_paths


def load_fgp_data(fgp_file):
    """Load data from the fgp (fam - genome - protein) tab delimited file from cazomevolve.

    :param fgp_file: str/Path, path to fgp_file

    Return pandas df with columns 'Family', 'Genome' and 'Protein' (containing the protein id/accession)
    """
    fgp_df = pd.read_table(fgp_file, header=None)
    fgp_df.columns = ['Family', 'Genome', 'Protein']

    return fgp_df


def add_tax_column_from_row_index(df, tax_rank, tax_index):
    """Add a column listing the tax for each row in a family frequency df
    where the tax info is in the row index
    
    :param df: df to add tax data to
    :param tax_rank: str, name of the tax rank, to be used as new column name
    :param tax_index: int, index of tax info in the row name/index to be added to the new column
    
    Return df with a new column add
    """
    new_col = []
    
    for i in range(len(df)):
        tax = df.iloc[i].name[tax_index]
        new_col.append(tax)
        
    df[tax_rank] = new_col
    
    return df


def load_tax_data(
    tax_csv_path,
    kingdom=False,
    phylum=False,
    tax_class=False,
    tax_order=False,
    tax_family=False,
    genus=False,
    species=False,
):
    """Load tax data compiled by cazomevolve into a pandas df

    :param tax_csv_path: str/Path to csv file of genome, tax_rank, tax_rank
        e.g. 'Genome', 'Genus', 'Species'
    The remaining params are bool checks for lineage ranks included in the tax data file
    
    Return df of genome, tax_rank, tax_rank,  e.g. 'Genome', 'Genus', 'Species'
    """
    tax_df = pd.read_csv(tax_csv_path)
    
    col_names = ['Genome']
    if kingdom:
        col_names.append('Kingdom')
    if phylum:
        col_names.append('Phylun')
    if tax_class:
        col_names.append('Class')
    if tax_order:
        col_names.append('Order')
    if tax_family:
        col_names.append('Tax_Family')
    if genus:
        col_names.append('Genus')
    if species:
        col_names.append('Species')

    try:
        tax_df = tax_df.drop('Unnamed: 0', axis=1)  #'Unnamed: 0' from pandas indexes
    except:
        pass

    for colname in tax_df.columns:
        if colname not in col_names:
            tax_df = tax_df.drop(colname, axis=1)

    return tax_df


def add_tax_data_from_tax_df(
    df,
    tax_df,
    kingdom=False,
    phylum=False,
    tax_class=False,
    tax_order=False,
    tax_family=False,
    genus=False,
    species=False,
):
    """Extract tax data from the tax df and add to the df (e.g. the gfp_df)
    
    :param df: pandas df, df to add tax data to
    :param df: pandas df containing tax data, with one column called 'Genome'
        and one column per tax rank
    The remaining params are bool checks for lineage ranks to be added to 
    the df
    
    Return df with new taxonomy columns
    """
    tax_ranks = []
    if kingdom:
        tax_ranks.append('Kingdom')
    if phylum:
        tax_ranks.append('Phylun')
    if tax_class:
        tax_ranks.append('Class')
    if tax_order:
        tax_ranks.append('Order')
    if tax_family:
        tax_ranks.append('Tax_Family')
    if genus:
        tax_ranks.append('Genus')
    if species:
        tax_ranks.append('Species')
    if len(tax_ranks) == 0:
        print('No tax ranks listed to be added to df')
        return df
    
    for tax_rank in tax_ranks:
        new_col = []
        for ri in tqdm(range(len(df)), desc=f"Collecting {tax_rank} data"):
            # retrieve the row in the tax_df containing the corresponding genome information
            tax_row = tax_df[tax_df['Genome'] == df.iloc[ri]['Genome']]
            new_col.append(tax_row[tax_rank].values[0])
        df[tax_rank] = new_col
        
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scripts in the script module

These test are intened to be run from the root of the repository using:
pytest -v
"""


import logging
import pytest
import numpy as np

from argparse import Namespace
from pathlib import Path

from cazomevolve.cazome.explore import explore_cazomes



@pytest.fixture
def argsdict(test_input_dir, test_output_dir):
    out = test_output_dir / "get_dbcan_cazymes"
    fgp_path = test_input_dir / "cazome_data_files/FGP_FILE"
    fg_path = test_input_dir / "cazome_data_files/FG_FILE"
    return {'args': Namespace(
        output_dir=out,
        dbcan_dir=test_input_dir,
        fam_genome_list=fg_path,
        fam_genome_protein_list=fgp_path,
        force=False,
        nodelete=True,
        input_dir=test_input_dir,
        dbcan_version=None,
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
    )}


def test_explore_main_no_tax(monkeypatch):
    def mock_run(*args, **kwards):
        return
    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_run)

    argsdict = {'args': Namespace(
        output_dir=Path("tests/test_outputs"),
        force=False,
        nodelete=True,
        dbcan_version=None,
        kingdom=False,
        phylum=False,
        tax_class=False,
        tax_order=False,
        tax_family=False,
        genus=False,
        species=False,
    )}

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        explore_cazomes.main(args=argsdict['args'], logger=logging.getLogger(__name__))
    assert pytest_wrapped_e.type == SystemExit


def test_get_dbcan_main(argsdict, monkeypatch):
    """Test the main entry point for build_cazy_db.py"""   
    def mock_none(*args, **kwards):
        return
    
    def mock_compare_fams(*args, **kwards):
        return 1,2,3

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none) 
    monkeypatch.setattr(explore_cazomes, "load_data", mock_none) 
    monkeypatch.setattr(explore_cazomes, "compare_cazome_sizes", mock_none) 
    monkeypatch.setattr(explore_cazomes, "compare_cazy_classes", mock_none) 
    monkeypatch.setattr(explore_cazomes, "compare_cazy_families", mock_compare_fams) # needs three 
    monkeypatch.setattr(explore_cazomes, "compare_core_cazomes", mock_none) 
    monkeypatch.setattr(explore_cazomes, "find_always_cooccurring_families", mock_none) 
    monkeypatch.setattr(explore_cazomes, "closing_message", mock_none) 

    explore_cazomes.main(args=argsdict['args'], logger=logging.getLogger(__name__))


def test_load_data(monkeypatch, fgp_df):
    def mock_df(*args, **kwards):
        return fgp_df

    monkeypatch.setattr(explore_cazomes, "load_fgp_data", mock_df)
    monkeypatch.setattr(explore_cazomes, "load_tax_data", mock_df)
    monkeypatch.setattr(explore_cazomes, "add_tax_data_from_tax_df", mock_df)

    args = Namespace(
        fgp_file="tests",
        tax_csv_path="tests",
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
    )

    explore_cazomes.load_data(args)


def test_compare_cazome_sizes(monkeypatch, fam_freq_df_with_tax, test_output_dir):
    def mock_none(*args, **kwards):
        return

    def mock_counts(*args, **kwards):
        return {}, fam_freq_df_with_tax    

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "count_items_in_cazome", mock_counts)

    out = test_output_dir / "explore_cazomes"

    args = Namespace(
        proteome_dir=None,
        output_dir=out,
        group_by='Genus',
        round_by=2,
    )

    assert explore_cazomes.compare_cazome_sizes(fam_freq_df_with_tax, args) is None


def test_compare_cazome_sizes_proteome(monkeypatch, fam_freq_df_with_tax, test_output_dir):
    def mock_none(*args, **kwards):
        return

    def mock_counts(*args, **kwards):
        return {}, fam_freq_df_with_tax    

    def mock_proteome(*args, **kwards):
        return {
            'Genus1': {
                'GCA_003382565.3': 10,
                }
        }

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "count_items_in_cazome", mock_counts)
    monkeypatch.setattr(explore_cazomes, "calc_proteome_representation", mock_counts)
    monkeypatch.setattr(explore_cazomes, "get_proteome_sizes", mock_proteome)

    out = test_output_dir / "explore_cazomes"

    args = Namespace(
        proteome_dir=None,
        output_dir=out,
        group_by='Genus',
        round_by=2,
    )

    assert explore_cazomes.compare_cazome_sizes(fam_freq_df_with_tax, args) is None


def test_compare_classes(fam_freq_df_with_tax, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)

    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        proteome_dir=None,
        output_dir=out,
        group_by='Genus',
        round_by=2,
    )

    explore_cazomes.compare_cazy_classes(fam_freq_df_with_tax, args)


def test_compare_families(fam_freq_df_with_tax, built_fam_freq_df, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_built_df(*args, **kwards):
        return built_fam_freq_df

    def mock_row_colours(*args, **kwards):
        return built_fam_freq_df, built_fam_freq_df

    def mock_get_group_specific_fams(*args, **kwards):
        return {'grp': [1, 2, 3]}, []

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "build_fam_freq_df", mock_built_df)
    monkeypatch.setattr(explore_cazomes, "build_family_clustermap", mock_none)
    monkeypatch.setattr(explore_cazomes, "build_row_colours", mock_row_colours)
    monkeypatch.setattr(explore_cazomes, "get_group_specific_fams", mock_get_group_specific_fams)

    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        proteome_dir=None,
        output_dir=out,
        group_by='Genus',
        round_by=2,
        kingdom=False,
        phylum=False,
        tax_class=False,
        tax_order=False,
        tax_family=False,
        genus=True,
        species=True,
        formats=['pdf'],
        show_plots=False,
    )

    explore_cazomes.compare_cazy_families(fam_freq_df_with_tax, args)


def test_core_cazome(built_fam_freq_df, all_families, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_core_cazome(*args, **kwards):
        return ['GH1', 'PL1']
    
    def mock_df(*args, **kwards):
        return built_fam_freq_df
    
    def mock_build_mean_df(*args, **kwards):
        return built_fam_freq_df, built_fam_freq_df

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "identify_core_cazome", mock_core_cazome)
    monkeypatch.setattr(explore_cazomes, "add_tax_column_from_row_index", mock_df)
    monkeypatch.setattr(explore_cazomes, "build_fam_mean_freq_df", mock_build_mean_df)

    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        output_dir=out,
        fgp_file="tests",
        tax_csv_path="tests",
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
        group_by='Genus',
        round_by=2,
    )

    explore_cazomes.compare_core_cazomes(
        built_fam_freq_df,
        built_fam_freq_df.set_index(['Genome','Genus','Species']),
        all_families,
        args,
    )


def test_cooccurring_fams(built_fam_freq_df, all_families, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_cooccurring(*args, **kwards):
        return {
            0: {'fams': {'CBM4', 'GH148'}, 'freqs': {8}},
            1: {'fams': {'CBM5', 'CBM50', 'GH23', 'GH3', 'GT2', 'GT51', 'GT9'},
            'freqs': {711}},
            2: {'fams': {'GH121', 'GH146'}, 'freqs': {1}},
            3: {'fams': {'GH127', 'GH15'}, 'freqs': {1}},
            4: {'fams': {'GH94', 'GT84'}, 'freqs': {309}},
        }

    def mock_upsetplot_membership(*args, **kwards):
        return []

    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "build_upsetplot", mock_none)
    monkeypatch.setattr(explore_cazomes, "add_upsetplot_grp_freqs", mock_none)
    monkeypatch.setattr(explore_cazomes, "get_upsetplot_grps", mock_none)
    monkeypatch.setattr(explore_cazomes, "build_upsetplot_matrix", mock_none)
    monkeypatch.setattr(explore_cazomes, "calc_cooccuring_fam_freqs", mock_cooccurring)
    monkeypatch.setattr(explore_cazomes, "add_to_upsetplot_membership", mock_upsetplot_membership)

    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        output_dir=out,
        fgp_file="tests",
        tax_csv_path="tests",
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
        group_by='Genus',
        round_by=2,
        formats=['pdf'],
    )

    explore_cazomes.find_always_cooccurring_families(
        built_fam_freq_df,
        built_fam_freq_df.set_index(['Genome', 'Genus', 'Species']),
        all_families,
        args,
    )


def test_pca(built_fam_freq_df, all_families, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_pca(*args, **kwards):
        return Namespace(explained_variance_ratio_=np.array([1,2,3,4,5])), 'pca'
    
    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "plot_explained_variance", mock_none)
    monkeypatch.setattr(explore_cazomes, "plot_scree", mock_none)
    monkeypatch.setattr(explore_cazomes, "plot_pcs", mock_none)
    monkeypatch.setattr(explore_cazomes, "perform_pca", mock_pca)


    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        output_dir=out,
        fgp_file="tests",
        tax_csv_path="tests",
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
        group_by='Genus',
        round_by=2,
        formats=['pdf'],
        show_plots=False,
    )

    explore_cazomes.run_pca(
        built_fam_freq_df,
        built_fam_freq_df.set_index(['Genome','Genus','Species']),
        all_families,
        args,
    )

def test_plot_pcs(built_fam_freq_df, all_families, test_output_dir, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    monkeypatch.setattr(explore_cazomes, "make_output_directory", mock_none)
    monkeypatch.setattr(explore_cazomes, "plot_pca", mock_none)
    monkeypatch.setattr(explore_cazomes, "plot_loadings", mock_none)

    out = test_output_dir / "explore_cazomes"
    args = Namespace(
        output_dir=out,
        fgp_file="tests",
        tax_csv_path="tests",
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
        group_by='Genus',
        round_by=2,
        formats=['pdf'],
        show_plots=False,
    )

    explore_cazomes.plot_pcs(
        (1,2),
        built_fam_freq_df.set_index(['Genome', 'Genus', 'Species']),
        'pca',
        'xscaled',
        out,
        args,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# ContactC                                    
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Parse output from dbCAN version 2"""


import logging
import pandas as pd
import re

from tqdm import tqdm
from typing import List, Optional

from saintBioutils.utilities.file_io import make_output_directory
from saintBioutils.utilities.file_io.get_paths import get_dir_paths

from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    # make output dir if necessary
    if str(args.fam_genome_list.parent) != ".":
        make_output_directory(args.fam_genome_list.parent, args.force, args.nodelete)

    if str(args.fam_genome_protein_list.parent) != ".":
        if str(args.fam_genome_list.parent) == str(args.fam_genome_protein_list.parent):
            make_output_directory(args.fam_genome_protein_list.parent, True, True)
        else:
            make_output_directory(args.fam_genome_protein_list.parent, args.force, args.nodelete)

    # get path to output directories from dbCAN
    output_dirs = get_dir_paths(args.dbcan_dir)

    for output_dir in tqdm(output_dirs, desc="Parsing dbCAN output dirs"):
        get_family_annotations(output_dir, args)

    closing_message('Get dbCAN CAZymes', args)


def get_family_annotations(output_dir, args):
    """Extract CAZy family annotations from the dbCAN output
    
    Extracted famly annotations are added to the tab delimited list

    :param output_dir: Path, path to output dir
    :param args: cmd-line args parser
    
    Return nothing"""
    logger = logging.getLogger(__name__)

    genomic_accession = output_dir.name
    
    fam_annotations = {}  # {protein accession: {fams}} -- a single protein can appear on multiple lines

    try:
        df = pd.read_table(output_dir/"overview.txt")
    except FileNotFoundError:
        logger.error(f"Could not find overview.txt file in {output_dir.name}\nSkipping output dir")
        return

    # drop rows were #ofTools = 1
    df = df[df['#ofTools'] != 1]
    
    for ri in tqdm(range(len(df)), desc=f"Parsing {output_dir.name}"):
        row = df.iloc[ri]

        protein_accession = row['Gene ID']

        hmmer_fams = get_tool_fams(row[1])
        hotpep_fams = get_tool_fams(row[2])
        diamond_fams = get_tool_fams(row[3])

        # get the fams at least two tools agreed upon
        dbcan_fams = get_dbcan_consensus(hmmer_fams, hotpep_fams, diamond_fams)

        if len(dbcan_fams) > 0:
            try:
                fam_annotations[protein_accession]
                for fam in dbcan_fams:
                    fam_annotations[protein_accession].add(fam)
            except KeyError:
                fam_annotations[protein_accession] = dbcan_fams
    
    with open(args.fam_genome_list, 'a') as fh:
        for protein_acc in tqdm(fam_annotations, desc="Adding fam-genome annotations to tab delim list"):
            protein_fams = fam_annotations[protein_acc]
            for fam in protein_fams:
                fh.write(f"{fam}\t{genomic_accession}\n")

    with open(args.fam_genome_protein_list, 'a') as fh:
        for protein_acc in tqdm(fam_annotations, desc="Adding fam-genome-protein annotations to tab delim list"):
            protein_fams = fam_annotations[protein_acc]
            for fam in protein_fams:
                fh.write(f"{fam}\t{genomic_accession}\t{protein_acc}\n")

    return


def get_tool_fams(tool_data):
    """Get predicted CAZy family annotations for a specific tool
    
    :param tool_data: str, output from prediction tool
    
    Return set of CAZy family predictions"""
    print(tool_data, '***')
    if tool_data == "-":
        return set()
    
    predicted_domains = tool_data.split("+")  # each predicted domain is separated b "+"

    fams = set()

    for domain in predicted_domains:
        domain = domain.split("(")[0].split("_")[0]  # drop CAZy subfamily
        if domain.startswith(("G", "P", "C", "A")):  # filter out EC numbers
            fams.add(domain)
    
    return fams


def get_dbcan_consensus(hmmer_fams, hotpep_fams, diamond_fams):
    """Get the fams at least two tools predicted

    Hotpep == eCAMI - does not matter which verion of the dbCAN was used
    
    :param hmmer_fams: set of CAZy family annotations
    :param hotpep_fams: set of CAZy family annotations
    :param diamond_fams: set of CAZy family annotations
    
    Return LIST
    """
    print(hmmer_fams, hotpep_fams, diamond_fams, 'RESULTS')
    hmmer_hotpep = hmmer_fams & hotpep_fams
    hmmer_diamond = hmmer_fams & diamond_fams
    hotpep_diamond = hotpep_fams & diamond_fams
    all_tools = hmmer_fams & diamond_fams & hotpep_fams

    dbcan_consensus = list(all_tools.union(hmmer_hotpep, hmmer_diamond, hotpep_diamond))
    
    return dbcan_consensus


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Test taxs module add_taxs.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import pandas as pd

from argparse import Namespace

from saintBioutils.utilities import logger

from cazomevolve.taxs import add_taxs


@pytest.fixture
def argsdict(test_input_dir):
    fgp_path = test_input_dir / "cazome_data_files/FGP_FILE-uneditable"
    fg_path = test_input_dir / "cazome_data_files/FG_FILE-uneditable"

    return {'args': Namespace(
        email='email',
        FGP_FILE=fgp_path,
        FG_FILE=fg_path,
        kingdom=True,
        phylum=True,
        tax_class=True,
        tax_order=True,
        tax_family=True,
        genus=True,
        species=True,
        gtdb=None,
    )}


@pytest.fixture
def col_names():
    return ['Genome', 'Kingdom', 'Genus', 'Species']


@pytest.fixture
def gtdb_path(test_input_dir):
    _path = test_input_dir / "gtdb/gtdb_data.tsv"
    return _path


def test_no_files(monkeypatch):
    """Test when no files are specified"""
    def mock_logger(*args, **kwards):
        return

    monkeypatch.setattr(logger, "config_logger", mock_logger)
    monkeypatch.setattr(add_taxs, "config_logger", mock_logger)

    args = Namespace(
        email='email',
        FGP_FILE=None,
        FG_FILE=None,
        kingdom=False,
        phylum=False,
        tax_class=False,
        tax_order=False,
        tax_family=False,
        genus=False,
        species=False,
    )

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        add_taxs.main(args=args)
    assert pytest_wrapped_e.type == SystemExit


def test_no_taxs(monkeypatch):
    """Test when no lineages are specified"""
    def mock_logger(*args, **kwards):
        return

    monkeypatch.setattr(logger, "config_logger", mock_logger)
    monkeypatch.setattr(add_taxs, "config_logger", mock_logger)

    args = Namespace(
        email='email',
        FGP_FILE='1',
        FG_FILE='2',
        kingdom=False,
        phylum=False,
        tax_class=False,
        tax_order=False,
        tax_family=False,
        genus=False,
        species=False,
    )

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        add_taxs.main(args=args)
    assert pytest_wrapped_e.type == SystemExit


def test_add_tax_main(argsdict, monkeypatch):
    def mock_return_none(*args, **kwards):
        return
    def mock_get_gtdb(*args, **kwards):
        return {}, ['1', '2']

    monkeypatch.setattr(logger, "config_logger", mock_return_none)
    monkeypatch.setattr(add_taxs, "config_logger", mock_return_none)
    monkeypatch.setattr(add_taxs, "load_gtdb_df", mock_return_none)
    monkeypatch.setattr(add_taxs, "add_gtdb_taxs", mock_get_gtdb)
    monkeypatch.setattr(add_taxs, "add_ncbi_taxs", mock_return_none)
    monkeypatch.setattr(add_taxs, "write_tab_lists", mock_return_none)
    monkeypatch.setattr(add_taxs, "write_out_csv", mock_return_none)
    monkeypatch.setattr(add_taxs, "closing_message", mock_return_none)

    add_taxs.main(args=argsdict['args'])


def test_load_gtdb_none(argsdict, col_names):
    """args.gtdb = None"""
    data = {'Genome': [], 'Kingdom': [], 'Genus': [], 'Species': []}
    df = pd.DataFrame(data)
    new_df = add_taxs.load_gtdb_df(col_names, argsdict['args'])
    assert list(new_df.columns) == list(df.columns)
    assert len(new_df) == len(df)


def test_load_gtdb(argsdict, col_names, gtdb_path):
    """args.gtdb = None"""
    data = {'Genome': [0]*249, 'Kingdom': [0]*249, 'Genus': [0]*249, 'Species': [0]*249}
    df = pd.DataFrame(data)

    argsdict['args'].gtdb = gtdb_path
    argsdict['args'].phylum = False
    argsdict['args'].tax_order = False
    argsdict['args'].tax_class = False
    argsdict['args'].tax_family = False

    new_df = add_taxs.load_gtdb_df(col_names, argsdict['args'])
    
    assert list(new_df.columns) == list(df.columns)
    assert len(new_df) == len(df)


def test_add_taxs(argsdict, col_names, test_input_dir):
    _path = test_input_dir / "gtdb/parsed_gtdb.csv"
    df = pd.read_csv(_path, index_col="Unnamed: 0")

    argsdict['args'].phylum = False
    argsdict['args'].tax_order = False
    argsdict['args'].tax_class = False
    argsdict['args'].tax_family = False

    out1, out2 = add_taxs.add_gtdb_taxs(df, col_names, argsdict['args'])
    assert out1 == {}
    assert out2 == {'GCA_003382565.3'}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Explore the sizes od CAZy class populations per genus"""


import pandas as pd
import numpy as np
import re

from tqdm import tqdm


CAZY_CLASSES = ['GH', 'GT', 'PL', 'CE', 'AA', 'CBM']


def calculate_class_sizes(fgp_df, grp, round_by=None):
    """Calculate the mean (+- SD) of CAZymes per CAZy class per grp.

    Num of CAZymes is the number of unique protein accessions.

    :param fgp_df: pandas df, columns = ['Family', 'Genome', 'Protein', 'Genus', 'Species']
    :param grp: str, tax rank to group genomes by, e.g. 'Genus' or 'Species'
    :param round_by: int, num of dp to round the mean and sd to, if None does not round

    Return
    * df columns ['CAZyClass', grp, 'MeanCazyClass', 'SdCazyClass', 'NumOfGenomes']
    * dict cazy_class_size_dict
    """
    cazy_class_size_dict = {}  # {class: {grp: {genome: {'proteins': set(protein id), 'numOfProteins': int}}}}

    # calculate total CAZymes per genome - used to calculate the percentage of the cazome
    # made up of each CAZy class
    cazome_sizes = {}  # {grp: {genome: {'cazymes':{ids/accs}}}}

    for ri in tqdm(range(len(fgp_df)), desc="Getting CAZy class sizes"):
        protein_acc = fgp_df.iloc[ri]['Protein']
        fam = fgp_df.iloc[ri]['Family']
        grp_name = fgp_df.iloc[ri][grp]
        genome = fgp_df.iloc[ri]['Genome']

        try:
            cazome_sizes[grp_name]
        except KeyError:
            cazome_sizes[grp_name] = {}

        try:
            cazome_sizes[grp_name][genome]['CAZymes'].add(protein_acc)
        except KeyError:
            cazome_sizes[grp_name][genome] = {'CAZymes': {protein_acc}}
        
        cazy_class = re.match('\D{2,3}', fam).group()
        
        try:
            cazy_class_size_dict[cazy_class]
        except KeyError:
            cazy_class_size_dict[cazy_class] = {}
            
        try:
            cazy_class_size_dict[cazy_class][grp_name]
        except KeyError:
            cazy_class_size_dict[cazy_class][grp_name] = {}
            
        try:
            cazy_class_size_dict[cazy_class][grp_name][genome]['proteins'].add(protein_acc)
        except KeyError:
            cazy_class_size_dict[cazy_class][grp_name][genome] = {'proteins': {protein_acc}}
            
    cazy_class_data = []
    grps = set(fgp_df[grp])

    for cazy_class in tqdm(CAZY_CLASSES, desc="Calculating CAZy class sizes"):
        
        for grp_name in grps:
            try:
                cazy_class_size_dict[cazy_class][grp_name]
            except KeyError:
                # cazy class is not in any genomes from the grp_name
                num_genomes = len(set(fgp_df[fgp_df[grp] == grp_name]['Genome']))  # sample size
                cazy_class_data.append(
                    [cazy_class, grp_name, 0, 0, 0, 0, num_genomes]
                )
                continue
                
            # get sample size, i.e. num of genomes
            num_genomes = len(list(cazy_class_size_dict[cazy_class][grp_name].keys()))
            
            # calculate the number of CAZyme in the class
            cazy_class_sizes = []
            for genome in cazy_class_size_dict[cazy_class][grp_name]:
                num_of_cazymes = len(cazy_class_size_dict[cazy_class][grp_name][genome]['proteins'])
                cazy_class_size_dict[cazy_class][grp_name][genome]['numOfProteins'] = num_of_cazymes
                cazy_class_sizes.append(num_of_cazymes)
                
            mean_cazy_class = np.mean(cazy_class_sizes)
            sd_cazy_class = np.std(cazy_class_sizes)

            if round_by is not None:
                mean_cazy_class = round(mean_cazy_class, round_by)
                sd_cazy_class = round(sd_cazy_class, round_by)

            # calculate the percentage of the CAZome represented by the CAZy class
            cazy_class_percentages = []
            for genome in cazy_class_size_dict[cazy_class][grp_name]:
                total_cazymes = len(cazome_sizes[grp_name][genome]['CAZymes'])
                num_class_cazymes = len(cazy_class_size_dict[cazy_class][grp_name][genome]['proteins'])
                percentage = (num_class_cazymes / total_cazymes) * 100
                cazy_class_percentages.append(percentage)
                
            mean_perc_cazy_class = np.mean(cazy_class_percentages)
            sd_perc_cazy_class = np.std(cazy_class_percentages)

            if round_by is not None:
                mean_perc_cazy_class = round(mean_perc_cazy_class, round_by)
                sd_perc_cazy_class = round(sd_perc_cazy_class, round_by)
            
            cazy_class_data.append(
                [cazy_class, grp_name, mean_cazy_class, sd_cazy_class, mean_perc_cazy_class, sd_perc_cazy_class, num_genomes]
            )

    col_names = ['CAZyClass', grp, 'MeanCazyClass', 'SdCazyClass', 'MeanClassPerc', 'SdClassPerc', 'NumOfGenomes']
    class_df = pd.DataFrame(cazy_class_data, columns=col_names)
    
    return class_df, cazy_class_size_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Test cazomes.get_cazy_cazymes.py

These test are intened to be run from the root of the repository using:
pytest -v
"""


import pytest
import pandas as pd

from argparse import Namespace
from pathlib import Path

from cazy_webscraper.sql import sql_orm
from saintBioutils.utilities import logger

from cazomevolve.cazome.cazy import get_cazy_cazymes


@pytest.fixture
def argsdict(test_input_dir, test_output_dir):
    out = test_output_dir / "get_cazy_cazymes"
    fgp_path = test_input_dir / "cazome_data_files/FGP_FILE"
    fg_path = test_input_dir / "cazome_data_files/FG_FILE"

    return {'args': Namespace(
        output_dir=out,
        fam_genome_list=fg_path,
        fam_genome_protein_list=fgp_path,
        database='test',
        force=False,
        nodelete=True,
        sql_echo=False,
        input_dir=test_input_dir,
    )}


def test_cazy_main_no_files(argsdict, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_files(*args, **kwards):
        return []
    
    monkeypatch.setattr(get_cazy_cazymes, "make_output_directory", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "get_db_connection", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "get_file_paths", mock_files)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        get_cazy_cazymes.main(args=argsdict['args'])
    assert pytest_wrapped_e.type == SystemExit


def test_cazy_main(argsdict, monkeypatch):
    def mock_none(*args, **kwards):
        return
    
    def mock_files(*args, **kwards):
        return [1,2,3]
    
    monkeypatch.setattr(get_cazy_cazymes, "make_output_directory", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "get_db_connection", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "get_file_paths", mock_files)
    monkeypatch.setattr(get_cazy_cazymes, "get_gbk_table_dict", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "get_cazy_annotations", mock_none)
    monkeypatch.setattr(get_cazy_cazymes, "closing_message", mock_none)

    get_cazy_cazymes.main(args=argsdict['args'])


def test_get_cazy_annotations_invalid(argsdict):
    fasta_path = Path("tests/FILE.faa")
    assert None == get_cazy_cazymes.get_cazy_annotations(fasta_path, {}, argsdict['args'], 'connection')


def test_get_cazy_annotations(test_input_dir, argsdict, db_path):
    fasta_path = test_input_dir / "cazome_explore/GCA_003382565.3.faa"
    gbk_table_dict = {'CAG72925.1': 1, 'CAG72927.1': 2}
    db_connection = sql_orm.get_db_connection(db_path, False, False)
    get_cazy_cazymes.get_cazy_annotations(
        fasta_path,
        gbk_table_dict,
        argsdict['args'],
        db_connection,
    )